// *******************************************************************
//  A string class  Implementation
//
//  A string class consistent with a subset of the standard C++
//  string class as defined in the draft ANSI standard.
// *******************************************************************

#include <string.h>
#include <assert.h>
#include "astring.h"

const int npos = -1;
const int MAX_LENGTH = 1024;    // Maximum size of an input string

//****************************************************************

astring::astring()

// Constructs an empty string

{
    myLength = 0;    
    myCapacity = 1;
    myCstring = new char[myCapacity];
    myCstring[0] = '\0';           // Appends the terminating null char
}

//****************************************************************

astring::astring(const char *s)

// Constructs a string object from a literal string s

{
    assert (s != 0);               // C-string not NULL?
    
    myLength = strlen(s);
    myCapacity = myLength + 1;     // make room for '\0'
    myCstring = new char[myCapacity];  
    strcpy(myCstring,s);
}   

//****************************************************************

astring::astring(const astring &str)

// Copy constructor

{
    myLength = str.length();
    myCapacity = myLength + 1;
    myCstring = new char[myCapacity];
    strcpy(myCstring, str.myCstring);
}

//****************************************************************

astring::~astring()

// Destructor

{
    delete[] myCstring;                // free memory
}

//****************************************************************

const astring &astring::operator= (const astring &rhs)

{
    if (this != &rhs) {          // Do not assign to itself
        if (myCapacity < rhs.length() + 1) {      // If more memory needed?
            delete[] myCstring;                   //   delete old string
            myCapacity = rhs.length() + 1;        //   allocate new memory
            myCstring = new char[myCapacity];
        }
        myLength = rhs.length();
        strcpy(myCstring, rhs.myCstring);
    }
    return *this;   
}

//****************************************************************

const astring &astring::operator= (const char *s)

// Assignment from a literal string

{
    int len = 0;          // length of newly constructed string
    assert(s != 0);       // make sure s non-NULL
    len = strlen(s);      // # of characters in string
    
    // Free old string if necessary
    
    if (myCapacity < len + 1) {
        delete[] myCstring;          // delete old string
        myCapacity = len + 1;        // add 1 for '\0'
        myCstring = new char[myCapacity];       
    }
    myLength = len;
    strcpy(myCstring, s);
    return *this;
}

//****************************************************************

const astring &astring::operator= (char ch)

// Assignment from character as though single char string

{
    if (myCapacity < 2) {
        delete [] myCstring;
        myCapacity = 2;
        myCstring = new char[myCapacity];
    }
    myLength = 1;
    myCstring[0] = ch;    // make string one character long
    myCstring[1] = '\0';
    return *this;
}

//****************************************************************

int astring::length() const

// Returns the number of chars in string

{
    return myLength;
}

//****************************************************************

int astring::find( char ch ) const

// Returns the position of the first occurrence of the character
//   ch within this string.  If ch does not occur in this
//   string, then returns npos.

{
    int k;
    for(k = 0;   k < myLength;   k++) {
        if (myCstring[k] == ch)
            return k;
    }
    return npos;
}

//****************************************************************

int astring::find (const astring &str)  const
// Returns the position of the first occurrence of the substring
//   str within this string.  If str does not occur in this
//   string, then returns npos.

{
    int len = str.length();
    int lastIndex = myLength - len;
    int k;

    for (k = 0;   k <= lastIndex;   k++)
        if (strncmp(myCstring + k, str.c_str(), len) == 0)
            return k;
    return npos;
}

//****************************************************************

astring astring::substr(int pos, int len) const

// Extracts and returns the substring of length len starting
//   at position pos

{
    if (pos < 0)    // start at front when pos < 0
        pos = 0;
    
    if (pos >= myLength) return ""; // Return an empty string
    
    int lastIndex = pos + len - 1;  // Position of last char in substring
    if (lastIndex >= myLength)
        lastIndex = myLength-1;
    
    astring result(*this);          // make sure enough space allocated
    
    int j, k;
    for (j = 0, k = pos;   k <= lastIndex;   j++, k++)
        result.myCstring[j] = myCstring[k];

    result.myCstring[j] = '\0';         // properly terminate C-string
    result.myLength = j;                // record length properly
    return result;
}

//****************************************************************

const char *astring::c_str() const

// Returns the pointer to an equivalent null-terminated string

{
    return myCstring;    
}

//****************************************************************

char &astring::operator[] (int k)

// Checks subscript range: 0 <= k < myLength
// Returns a reference to the k-th character

{
    if (k < 0 || myLength <= k) {
        cerr << "index out of range: " << k << " string: " << myCstring
            << endl;
        assert (0 <= k && k < myLength);
    }
    return myCstring[k];
}     

//****************************************************************

char astring::operator[] (int k) const

// Checks subscript range: 0 <= k < myLength
// Returns the k-th character

{
    if (k < 0 || myLength <= k) {
        cerr << "index out of range: " << k << " string: " << myCstring
            << endl;
        assert(0 <= k && k < myLength);
    }
    return myCstring[k];
}

//****************************************************************

const astring &astring::operator+= (const astring &str)

// Concatenates a copy of str onto this string     

{

    astring copystring(str);         // Copy to avoid aliasing problems
    
    int newLength = myLength + str.length();
    int lastLocation = myLength;     // Position of the terminating '\0'
    
    // If local buffer not big enough:
    if (newLength >= myCapacity) {
        myCapacity = newLength + 1;
        char *newBuffer = new char[myCapacity];
        strcpy (newBuffer, myCstring); // Copy into new buffer
        delete [] myCstring;           // Delete old string
        myCstring = newBuffer;
    }
    
    // Now catenate copystring to end of myCstring:
    strcpy (myCstring+lastLocation, copystring.c_str());
    myLength = newLength;

    return *this;
}

//****************************************************************

const astring &astring::operator+= (char ch)

// Concatenates ch onto this string

{
    astring temp;    // make string equivalent of ch
    temp = ch;
    *this += temp;
    return *this;
}

//****************************************************************

ostream &operator<< (ostream &os, const astring &str)

// Outputs str to os

{
    return os << str.c_str();
}

//****************************************************************

istream &operator>> (istream &is, astring &str)

// Reads str from is

{
    char buf[MAX_LENGTH];
    is >> buf;
    str = buf;
    return is;
}

//****************************************************************

istream &getline (istream &is, astring &str)

// Reads a line from is into the string str

{
    char buf[MAX_LENGTH];
    is.getline(buf, MAX_LENGTH);
    str = buf;
    return is;
}

//****************************************************************

astring operator+ (const astring &lhs, const astring &rhs)

// Returns concatenation of lhs and rhs     

{
    astring result(lhs);   // copy lhs to result
    result += rhs;         // catenate rhs
    return result;         // return a copy of result
}

astring operator+ (char ch, const astring &str)

// Returns concatenation of ch and str

{
    astring result;    // make string equivalent of ch
    result = ch;
    result += str;
    return result;
}

astring operator+ (const astring &str, char ch)

// Returns concatenation of str with ch

{
    astring result(str);
    result += ch;
    return result;
}

//****************************************************************

bool operator== (const astring &lhs, const astring &rhs)

{
    return strcmp(lhs.c_str(), rhs.c_str()) == 0;
}

bool operator!= (const astring &lhs, const astring &rhs)

{
    return !(lhs == rhs);
}

bool operator< (const astring &lhs, const astring &rhs)

{
    return strcmp(lhs.c_str(), rhs.c_str()) < 0;
}

bool operator<= (const astring &lhs, const astring &rhs)

{
    return strcmp(lhs.c_str(), rhs.c_str()) <= 0;
}

bool operator> (const astring &lhs, const astring &rhs)

{
    return rhs < lhs;
}

bool operator>= (const astring &lhs, const astring &rhs)

{
    return rhs <= lhs;
}
