// MUSIC.CPP
//
// Music functions
//
// Author: Franz Liszt
//

#include <iostream.h>
#include "music.h"
#include "astack.h"

static int frequency[28] = {

//   C    D    E    F    G    A    B          Index
   131, 147, 165, 175, 196, 220, 247,     //  0 -  6
   262, 294, 330, 349, 392, 440, 494,     //  7 - 13
   523, 587, 659, 698, 784, 880, 988,     // 14 - 20 (Middle C = 14)
  1047,1175,1319,1397,1568,1760,1976      // 21 - 27
};

void PlayNote (NOTE &note, float tempo);
// Plays the note.  "note.tone" must be between 0 and 27
//    (Middle C = 14).
//   Halftones are not implemented.
//   note.tone = -1 indicates a pause without sound.
//   "note.duration" must be positive and is measured in 32-nds
//   (32 is a whole note, 16 is 1/2, etc.).
//   The actual duration is (tempo * note.duration / 32) seconds;

//****************************************************************

void PlayMusicSimple (MUSICELMT music[], float tempo)

// Plays notes from the music array.  Only elements of type
//   PLAYNOTE are considered, other elements are ignored.

{
    for (int i = 0;   music[i].type != PLAYSTOP;   i++)
        if (music[i].type == PLAYNOTE)
            PlayNote(music[i].note, tempo);
}

//****************************************************************

void PlayMusic (MUSICELMT music[], float tempo)

// Plays notes and fragments from the music array.
//   A fragment is played until the program reaches its finish or
//   a PLAYSTOP element.  The length of the initial fragment
//   is unknown: it must end with a PLAYSTOP element.

{
    STACK stack;
    PLAY type;   // Will hold the type of the current MUSICELMT

    // Create a stack for 400 integers:
    if (Create(stack, 400) == FAILED) {
        cerr << "*** MUSIC: Stack allocation error. ***\n";
        return;
    }

    int current = 0;      // Index of the current element.
    int finish = 9999;    // Index of the last element in a fragment
                          //   initially set to an arbitrary large
                          //   number.

    while (OK) {          // Repeat until break

        type = music[current].type;

        // If more elements remain in this fragment...
        if (current <= finish && type != PLAYSTOP) {
            if (type == PLAYNOTE)
                PlayNote(music[current++].note, tempo);
            else if (type == PLAYFRAGMENT) {
                // Save next and final element indices in the current
                //   fragment on stack:
                ...
                ...
                // Set current to the start of the new fragment and
                //  finish to the finish of the new fragment.
                ...
                ...
            }
        }
        // ...else if some interrupted fragments remain on stack...
        else if (!IsEmpty(stack)) {
            // Return to the last interrupted fragment
            ...
            ...
        }
        else
        // Nothing else to do
            break;
    }
    Destroy(stack);
}

//****************************************************************
//**********   PlayNote -- machine-dependent code      ***********
//****************************************************************

#ifdef _Windows   // MS Windows code

#include <windows.h>

void PlayNote (NOTE &note, float tempo)

{
    int msec = (int)(tempo * note.duration / 32 * 1000.);
    int ticks = (int)(tempo * note.duration / 32 * 380.);

    if (note.tone >= 0) {
        OpenSound();
        SetVoiceSound(1,(long)frequency[note.tone]*65536L, ticks);
        StartSound();
        WaitSoundState(S_QUEUEEMPTY);
        CloseSound();
    }
    else {
        MSG msg;
        WORD timer = SetTimer(NULL, 0, msec, NULL);
        while (!GetMessage(&msg, NULL, WM_TIMER, WM_TIMER));
        KillTimer(NULL, timer);
    }
}

//****************************************************************

#else
#ifdef __MSDOS__   // MS DOS code

#include <dos.h>      // Contains prototypes for
                      //   sound(freq), delay(ms), nosound()

void PlayNote (NOTE &note, float tempo)

{
    int msec = (int)(tempo * note.duration / 32 * 1000.);

    if (note.tone >= 0) {
        sound(frequency[note.tone]);  // Speaker on.
        delay (msec-10);    // Suspend for (msec-10) milliseconds.
        nosound();          // Speaker off.
        delay (10);         // Separate notes by a 10 msec break.
    }
    else
        delay (msec);       // tone < 0 indicates a pause.
}

#else

//****************************************************************

// Text output

void PlayNote (NOTE &note, float tempo)

{
    int msec = (int)(tempo * note.duration / 32 * 1000.);
    cout << "tone = " << note.tone << " duration = " << msec << " msec\n";
}

#endif
#endif
