// STR_CLS.CPP
//
// STRING class implementation

#include <ctype.h>
#include "str_cls.h"

static const int MAXLINE = 512;  // Max number of chars for ReadLine, >>

//****************************************************************

STRING::STRING()

// Default constructor.

{
    length = 0;
    buf = 0;
}

//****************************************************************

STRING::STRING (int a_length, char ch)

// Constructor from length with fill ch (default ' ').

{
    length = 0;
    buf = 0;
    if (a_length > 0 && a_length <= MAXLENGTH) {
        buf = new char[a_length+1];
        if (buf) {
            length = a_length;
            for (int i = 0;   i < length;   i++) buf[i] = ch;
            buf[length] = '\0';
        }
        else
            Error ("Constructor: memory allocation error");
    }
}

//****************************************************************

STRING::STRING (const STRING &str)

// Copy constructor.

{
    length = 0;
    buf = 0;
    int newlength = str.length;
    if (newlength > 0) {
        buf = new char[newlength+1];
        if (buf) {
            length = newlength;
            strcpy(buf, str.buf);
        }
        else
            Error ("Constructor: memory allocation error");
    }
}

//****************************************************************

STRING::STRING (const char *sz)

// Constructor from a null-term string.

{
    length = 0;
    buf = 0;
    int newlength = strlen(sz);
    if (newlength > 0 && newlength <= MAXLENGTH) {
        buf = new char[newlength+1];
        if (buf) {
            length = newlength;
            strcpy(buf, sz);
        }
        else
            Error ("Constructor: memory allocation error");
    }
}

//****************************************************************

STRING::~STRING ()

// Destructor.

{
    delete [] buf;
}

const STRING &
STRING::operator= (const STRING &str)

// Overloaded assignment.

{
    if (&str == this) return *this;

    if (length != str.length) {
        delete [] buf;
        length = 0;
        buf = 0;
        if (str.length > 0) {
            buf = new char[str.length+1];
            if (buf)
                length = str.length;
            else                
                Error ("Assignment: memory allocation error");
        }
    }
    if (buf)
        strcpy(buf, str.buf);

    return *this;
}

//****************************************************************

const STRING &
STRING::operator= (const char *sz)

// Automatic conversion from char *.

{
    if (buf != sz) {
        int newlength = strlen(sz);
        if (length != newlength && newlength <= MAXLENGTH) {
            delete [] buf;
            length = 0;
            buf = 0;
            if (newlength > 0) {
                buf = new char[newlength+1];
                if (buf)
                    length = newlength;
                else
                    Error ("Assignment: memory allocation error");
            }
        }
        if (buf)
            strcpy(buf, sz);
    }
    return *this;
}

//****************************************************************

int
STRING::Find (const char ch) const

// Returns the position of the first occurrence of ch
//   or NPOS if not found.

{
    char *p;

    if (buf && (p = strchr(buf, ch)) != 0)
        return int(p - buf);
    else
        return NPOS;
}

//****************************************************************

int
STRING::Find (const STRING &substr) const

// Returns the position of the first occurrence of the substring
//   or NPOS if not found.

{
    char *p;

    if (buf && (p = strstr(buf, substr.buf)) != 0)
        return int(p - buf);
    else
        return NPOS;
}

//****************************************************************

STRING
STRING::Substring (int pos, int len) const

// Returns the substring of len chars starting at pos.

{
    STRING str;

    if (len <= 0 || pos < 0 || pos > length) {
        Error ("Substring: indexing error");
        return str;
    }
    if (pos + len > length)
        len = length - pos;

    char *sz = new char [len+1];
    if (!sz) {
        Error ("Substring: memory allocation error");
        return str;
    }

    strncpy(sz, buf+pos, len);
    sz[len] = '\0';
    str.length = len;
    str.buf = sz;
    return str;
}

//****************************************************************

const STRING &
STRING::operator+= (const STRING &str)

// Overloaded += operator appends a string.

{
    if ((unsigned)length + (unsigned)str.length >=
                                      (unsigned)MAXLENGTH) {
        Error ("Concatenation: maximum length exceeded");
        return *this;
    }
    int newlength = length + str.length;
    char *sz = new char[newlength+1];
    if (!sz) {
        Error ("Concatenation: memory allocation error");
        return *this;
    }
    strcpy(sz, buf);
    strcpy(sz+length, str.buf);
    delete [] buf;
    length = newlength;
    buf = sz;

    return *this;
}

//****************************************************************

const STRING &
STRING::operator+= (const char ch)

// Overloaded += operator appends a char.

{
    if (length + 1 >= MAXLENGTH) {
        Error ("Concatenation: maximum length exceeded");
        return *this;
    }
    int newlength = length + 1;
    char *sz = new char[newlength+1];
    if (!sz) {
        Error ("Concatenation: memory allocation error");
        return *this;
    }
    strcpy(sz, buf);
    sz[length] = ch;
    sz[newlength] = '\0';
    delete [] buf;
    length = newlength;
    buf = sz;

    return *this;
}

//****************************************************************

int
STRING::ReadLine (istream &inp)

// Reads a line (up to MAXLINE chars) into string.
// Returns the new length of the string.

{
    delete [] buf;
    length = 0;
    char *sz = new char [MAXLINE+1];
    if (!sz) {
        Error ("ReadLine: memory allocation error");
        return length;
    }
    inp.getline(sz, MAXLINE+1, '\n');
    int newlength = strlen(sz);
    if (newlength > 0) {
        buf = new char[newlength+1];
        if (buf) {
            length = newlength;
            strcpy (buf, sz);
        }
        else
            Error ("ReadLine: memory allocation error");
    }
    delete [] sz;
    return length;
}

//****************************************************************
//***  Global operators and functions for the STRING class:    ***
//****************************************************************

STRING operator+ (const STRING &str1, const STRING& str2)

// Returns a new string equal to concatenated str1 and str2.

{
    STRING str;

    if ((unsigned)str1.length + (unsigned)str2.length >=
                                           (unsigned)MAXLENGTH) {
        STRING::Error ("Concatenation: maximum length exceeded");
        return str;
    }
    int newlength = str1.length + str2.length;
    char *sz = new char[newlength+1];
    if (!sz) {
        STRING::Error ("Concatenation: memory allocation error");
        return str;
    }
    strcpy(sz, str1.buf);
    strcpy(sz+str1.length, str2.buf);
    str.length = newlength;
    str.buf = sz;

    return str;
}

//****************************************************************

STRING operator+ (const char ch, const STRING &str2)

// Returns a new string equal to concatenated ch and str2.

{
    STRING str;

    if (1 + str2.length >= MAXLENGTH) {
        STRING::Error ("Concatenation: maximum length exceeded");
        return str;
    }
    int newlength = 1 + str2.length;
    char *sz = new char[newlength+1];
    if (!sz) {
        STRING::Error ("Concatenation: memory allocation error");
        return str;
    }
    sz[0] = ch;
    strcpy(sz+1, str2.buf);
    str.length = newlength;
    str.buf = sz;

    return str;
}

//****************************************************************

STRING operator+ (const STRING &str1, const char ch)

// Returns a new string equal to concatenated str1 and ch.

{
    STRING str;

    if (str1.length + 1>= MAXLENGTH) {
        STRING::Error ("Concatenation: maximum length exceeded");
        return str;
    }
    int newlength = str1.length + 1;
    char *sz = new char[newlength+1];
    if (!sz) {
        STRING::Error ("Concatenation: memory allocation error");
        return str;
    }
    strcpy(sz, str1.buf);
    sz[str1.length] = ch;
    sz[newlength] = '\0';
    str.length = newlength;
    str.buf = sz;

    return str;
}

//****************************************************************

istream &operator>> (istream &inp,  STRING &str)

// Stream extraction operator reads a word (up to MAXLINE chars)
//   into string.

{
    delete [] str.buf;
    str.length = 0;
    char *sz = new char [MAXLINE+1];
    if (!sz) {
        STRING::Error ("Stream >>: memory allocation error");
        return inp;
    }
    int newlength = 0;
    while (isspace(inp.peek())) inp.get(); // Skip whitespace
    while (newlength < MAXLINE) {
        if (isspace(inp.peek())) break;
        sz[newlength++] = inp.get();
    }
    sz[newlength] = '\0';
    if (newlength > 0) {
        str.buf = new char[newlength+1];
        if (str.buf) {
            str.length = newlength;
            strcpy (str.buf, sz);
        }
        else
            STRING::Error ("ReadLine: memory allocation error");
    }
    delete [] sz;
    return inp;
}

//****************************************************************

void
STRING::Error (const char *msg)

{
    cout << "*** STRING class error " << msg << " ***\n";
}

//****************************************************************


