// STR_CLS.H
//
// STRING class definition

#ifndef _STRING_CLS_H_
#define _STRING_CLS_H_

#include <iostream.h>
#include <string.h>
#include "bool.h"

static const int NPOS = -1;  // Returned by Find(...) when a position
                             //   in the string is not found.

static const int MAXLENGTH = 32000;
                             // Maximum length of a string.

class STRING {

  public:

    STRING ();              // Default constructor.
    STRING (int a_length, char ch = ' ');
                            // Constructor for a given length, fills
                            //   string with ch (default = ' ').
    STRING (const STRING &str);
                            // Copy constructor.
    STRING (const char *sz);
                            // Constructor from a null-term'ed string.
    ~STRING ();             // Destructor.

    const STRING &operator= (const STRING &str);
                            // Overloaded assignment.
    const STRING &operator= (const char *sz);
                            // Automatic conversion from char *.

    inline int Length () const;
                            // Returns the string length.
    inline char &operator[] (int i);
                            // Overloaded subscript operator [] -- works
                            //   in lvalue as well as in rvalue.

    inline operator const char * () const;
                            // Overloaded cast to const char *
                            //   operator converts a string into
                            //   the null-terminated string buffer.

    int Find (const char ch) const;
                            // Returns the position of the first
                            //   occurrence of ch (or NPOS).
    int Find (const STRING &substr) const;
                            // Returns the position of the first
                            //   occurrence of substr (or NPOS).
    STRING Substring (int pos, int len) const;
                            // Returns the substring of len chars
                            //   starting at pos.

    const STRING &operator+= (const STRING &str);
                            // Overloaded += operator appends a string.
    const STRING &operator+= (const char ch);
                            // Overloaded += operator appends a char.
    int ReadLine (istream &inp);
                            // Reads one line from inp and returns the
                            //   length of the new string.

    // *** Global operators and functions for the STRING class: ***

    // Concatenation:
    friend STRING operator+ (const STRING &str1, const STRING &str2);
    friend STRING operator+ (const char ch, const STRING &str2);
    friend STRING operator+ (const STRING &str1, const char ch);

    // Relational operators:
    friend bool operator== (const STRING &str1, const STRING &str2);
    friend bool operator!= (const STRING &str1, const STRING &str2);
    friend bool operator<  (const STRING &str1, const STRING &str2);
    friend bool operator<= (const STRING &str1, const STRING &str2);
    friend bool operator>  (const STRING &str1, const STRING &str2);
    friend bool operator>= (const STRING &str1, const STRING &str2);

    // Stream I/O:
    friend istream &operator>> (istream &inp,  STRING &str);
    friend ostream &operator<< (ostream &outp, const STRING &str);

  private:     // ... or "protected" for derived classes

    char *buf;
    int length;

    static void Error (const char *msg);
};

//****************************************************************
//*********               Inline functions              **********
//****************************************************************

inline int
STRING::Length () const

// Returns the string length.

{
    return length;
}

inline char &
STRING::operator[] (int i)

// Overloaded index operator [] -- works in lvalue as well as in rvalue.

{
    if (i < 0 || i >= length) {
        Error ("Index out of range");
        return buf[0];
    }
    return buf[i];
}

inline STRING::operator const char * () const

// Overloaded cast to const char * operator converts the string
//   into the constant pointer to the null-terminated string buffer.

{
    return buf;
}

            //*** Relational operators (based on strcmp): ***

inline bool operator== (const STRING &str1, const STRING &str2)

{
    return (strcmp(str1.buf, str2.buf) == 0);
}

inline bool operator!= (const STRING &str1, const STRING &str2)

{
    return (strcmp(str1.buf, str2.buf) != 0);
}

inline bool operator<  (const STRING &str1, const STRING &str2)

{
    return (strcmp(str1.buf, str2.buf) < 0);
}

inline bool operator<= (const STRING &str1, const STRING &str2)

{
    return (strcmp(str1.buf, str2.buf) <= 0);
}

inline bool operator>  (const STRING &str1, const STRING &str2)

{
    return (strcmp(str1.buf, str2.buf) > 0);
}

inline bool operator>= (const STRING &str1, const STRING &str2)

{
    return (strcmp(str1.buf, str2.buf) >= 0);
}

            //*** Stream I/O insertion: ***


inline ostream &operator<< (ostream &outp, const STRING &str)

// Stream insertion operator.

{
    return outp << str.buf;
}

#endif  // _STRING_CLS_H

