// TREE.CPP

typedef char SOMETYPE[25];
#include "tree.h"

//****************************************************************

void Destroy(TREENODE *root)

{
    // Base case 0: the tree is empty -- do nothing;
    if (root) {  // Recursive case:
        Destroy(root->left);
        Destroy(root->right);
        delete root;
    }
}

//****************************************************************
#ifdef RECURSIVE

TREENODE *Find(TREENODE *root, const SOMETYPE data)

// Finds the node such that node->data == data.
// Returns the pointer to the found node, or 0 if not found.

{
    if (!root)     // Base case 0: the tree is empty.
        return 0;

    int comp = Compare(data, root->data);

    if (comp == 0) // Base case: element found.
        return root;

    else if (comp < 0)                  // Recursive case: search
        return Find(root->left, data);  //   left subtree.

    else // if (comp > 0)               // Recursive case: search
        return Find(root->right, data); //   right subtree.
}

#else

TREENODE *Find(TREENODE *root, const SOMETYPE data)

// *** Iterative version ***

{
    while (root) {

        int comp = Compare(data, root->data);

        if (comp == 0)
            break; // element found
        if (comp < 0)
            root = root->left;
        else // if (comp > 0)
            root = root->right;
    }
    return root;
}

#endif

//****************************************************************
#ifdef RECURSIVE

STATUS Insert(TREENODE* &root, const SOMETYPE data)

// Creates a new node with node->data = data and inserts
//   it into the proper place in the tree.
// Returns OK if successful, FAILED if cannot
//    allocate a new node.

{
    if (!root) { // Base case 0: the tree is empty.

        // Allocate a new node and set the root
        //   equal to the new node:

        TREENODE *newnode = new TREENODE;
        if (!newnode) return FAILED;

        Copy (data, newnode->data);
        newnode->left =
        newnode->right = 0;
        root = newnode;
        return OK;
    }

    int comp = Compare(data, root->data);

    if (comp == 0) // Base case: the element is already in the tree.
        return OK; //   Another version might report an error.

    else if (comp < 0)                    // Recursive case: insert
        return Insert(root->left, data);  //   into the left subtree.

    else // if (comp > 0)                 // Recursive case: insert
        return Insert(root->right, data); //   into the right
                                          //   subtree.
}

#else

STATUS Insert(TREENODE* &root, const SOMETYPE data)

// *** Iterative version ***

{
	TREENODE **pNode = &root; // Pointer to a pointer to a TREENODE...
    while (*pNode) {
        int comp = Compare(data, (*pNode)->data);
        if (comp == 0) // The element already in the tree.
            return OK; //   Another version might report an error.
        if (comp < 0)
            pNode = &((*pNode)->left);
        else // if (comp > 0)
            pNode = &((*pNode)->right);
    }

    // Allocate a newnode and set data

    TREENODE *newnode = new TREENODE;
    if (!newnode) return FAILED;

    Copy (data, newnode->data);
    newnode->left =
    newnode->right = 0;

    // Append the node to the tree
    *pNode = newnode;

    return OK;
}

#endif

//****************************************************************
#ifdef RECURSIVE

STATUS Remove(TREENODE* &root, const SOMETYPE data)

// Removes the node whose data element is equal to data.
// Returns OK if successful, FAILED if cannot
//    find the node.

{
    if (!root) // Base case 0: the tree is empty.
        return FAILED;

    int comp = Compare(data, root->data);

    if (comp == 0) { // Base case:
                     //   found the node. root refers to
                     //   the pointer to the node to be removed...

        // Case 1: root has no more than one child. Replace root
        //   pointer with the pointer to that child.
        //   That unlinks the root node from the tree.
        //   Then delete the node.

        // First save root -- it points to the
        //   node to be removed:

        TREENODE *oldroot = root;

        if (root->left == 0)       // No left child
            root = root->right;
        else if (root->right == 0) // No right child
            root = root->left;
        if (oldroot != root) {     // If one of the above...
            delete oldroot;
            return OK;
        }

        // Case 2: root has both children.
        //   Find the smallest element 
        //   in the right subtree.  Unlink that element and
        //   make it the new root.

        // First find the smallest element in the right subtree.
        //   To do that, keep going left as long as possible.
        //   Also keep track of its parent

        TREENODE *parent = root;
        TREENODE *next = root->right;
        while (next->left) {
            parent = next;
            next = next->left;
        }

        // Unlink this node from the tree (it doesn't have
        //   a left child, so we can connect its right child to its
        //   parent)

        if (parent == root)
            root->right = next->right;
        else
            parent->left = next->right;

        // Make this element the new root

        next->left = root->left;
        next->right = root->right;
        delete root;
        root = next;

        return OK;
    }

    else if (comp < 0)                    // Recursive case: remove
        return Remove(root->left, data);  //   from the left subtree

    else // if (comp > 0)                 // Recursive case: remove
        return Remove(root->right, data); //   from the right subtree
}

#else

STATUS Remove(TREENODE* &root, const SOMETYPE data)

// *** Iterative version ***

{
    ...
    ...
}

#endif
